package
{
	import flash.display.*;
	import flash.external.*;
	import flash.events.*;
	import flash.ui.Keyboard;
	import mx.events.*;

	/**
	 * Game field
	 */
	public class Field extends EventDispatcher
	{
		public static const COMMAND_RIGHT:String = "right";
		public static const COMMAND_LEFT:String = "left";
		public static const COMMAND_DOWN:String = "down";
		public static const COMMAND_CWR:String = "cwr"; // clockwise rotation
		public static const COMMAND_CCWR:String = "ccwr"; // counterclockwise rotation

		public static const STATE_STOP:String = "stop";
		public static const STATE_FALLING:String = "falling";
		public static const STATE_DROPPING:String = "dropping";
		public static const STATE_DISAPPEARING:String = "disapearing";

		public static const WIDTH:int = 10;
		public static const HEIGHT:int = 20;

		[Bindable]
		/**
		 * level
		 */
		public var level:int;
		private var fallCount:int;

		private var _state:String;

		[Bindable(event=PropertyChangeEvent)]
		/**
		 * Get current state.
		 */
		public function get state():String
		{
			return _state;
		}

		/**
		 * playing or not
		 */
		public function get playing():Boolean
		{
			return state != STATE_STOP;
		}

		/**
		 * wait time
		 */
		private function get fallCountThreshold():int
		{
			return Math.max(0, 10 - level) * 2;
		}
		
		public var boxes:SimpleMatrix;
		public var tetrimino:Tetrimino;
		
		/**
		 * constructor
		 */
		public function Field(stage:Stage)
		{
			_state = STATE_STOP;
			level = 0;
			boxes = new SimpleMatrix();

			// initialize event
			stage.addEventListener("enterFrame", onEnterFrame);
			stage.addEventListener("keyDown", onKeyDown);
		}

		/**
		 * change current state
		 */
		private function setState(value:String):void
		{
			if(value != state)
			{
				var old:String  = state;
				_state = value;
				dispatchEvent(PropertyChangeEvent.createUpdateEvent(this, "state", old, value));
			}
		}

		/**
		 * start game
		 */
		public function play():void
		{
			if(!playing)
			{
				setState(STATE_FALLING);
			}
		}

		/**
		 * pause game
		 */
		public function pause():void
		{
			if(playing)
			{
				setState(STATE_STOP);
			}

			trace(state);
			trace(playing);
		}

		/**
		 * enterFrame event handler
		 */
		private function onEnterFrame(event:Event):void
		{
			if(tetrimino == null)
			{
				setNextTetrimino();
			}

			if(playing)
			{
				fallCount++;
				if(fallCount >= fallCountThreshold)
				{
					fallCount = 0;
					DoCommand(Field.COMMAND_DOWN);
				}
			}
		}

		/**
		 * keyDown event handler
		 */
		private function onKeyDown(e:KeyboardEvent):void
		{
			if(!playing)
			{
				return;
			}

			switch(e.keyCode)
			{
				case Keyboard.RIGHT:
					DoCommand(Field.COMMAND_RIGHT);
					break;

				case Keyboard.LEFT:
					DoCommand(Field.COMMAND_LEFT);
					break;

				case Keyboard.DOWN:
					DoCommand(Field.COMMAND_DOWN);
					break;

				case 88: // X
					DoCommand(Field.COMMAND_CCWR);
					break;
				case 90: // Y
					DoCommand(Field.COMMAND_CWR);
					break;
			}
		}

		public function DoCommand(command:String):void
		{
			if(!playing || tetrimino == null)
			{
				return;
			}

			var t:Tetrimino = tetrimino.clone();
			switch(command)
			{
				case COMMAND_RIGHT:
					t.x++;
					moveIfPossible(t);
					break;

				case COMMAND_LEFT:
					t.x--;
					moveIfPossible(t);
					break;

				case COMMAND_DOWN:
					t.y++;
					if(!moveIfPossible(t))
					{
						setTetriminoToBoxes();
					}
					break;

				case COMMAND_CWR:
					t.rotate(true);
					moveIfPossible(t);
					break;

				case COMMAND_CCWR:
					t.rotate(false);
					moveIfPossible(t);
					break;
			}
		}

		private function moveIfPossible(t:Tetrimino):Boolean
		{
			if(!tetrimino || !t)
			{
				return false;
			}

			for each(var box:Object in t.boxes)
			{
				var newX:int = t.x + box.x;
				var newY:int = t.y + box.y;

				if(newX >= 0 && newX < WIDTH && newY >= 0 && newY < HEIGHT
				&& boxes.getAt(newX, newY) == null)
				{
					continue;
				}
				return false;
			}

			tetrimino.x = t.x;
			tetrimino.y = t.y;
			tetrimino.angle = t.angle;
			return true;
		}

		private function setTetriminoToBoxes():void
		{
			// save drop point
			for each(var box:Object in tetrimino.boxes)
			{
				var newX:int = tetrimino.x + box.x;
				var newY:int = tetrimino.y + box.y;

				boxes.setAt(newX, newY, tetrimino.kind);
			}

			// clear tetrimino
			tetrimino = null;

			// line disapear check
			var disapearList:Array = getFullLine();
			if(disapearList.length > 0)
			{
				var event:DynamicEvent = new DynamicEvent("disapear");
				event.line = disapearList;
				dispatchEvent(event);

				var targetY:int = HEIGHT - 1;
				for(var y:int = HEIGHT - 1; y >= 0; y--)
				{
					if(disapearList.indexOf(y) != -1)
					{
						continue;
					}

					for(var x:int = 0; x < WIDTH; x++)
					{
						boxes.setAt(x, targetY, boxes.getAt(x, y));
					}
					targetY--;
				}

				for(y = targetY; y >= 0; y--)
				{
					for(x = 0; x < WIDTH; x++)
					{
						boxes.setAt(x, y, null);
					}
				}
			}
		}

		/**
		 * get full list
		 */
		private function getFullLine():Array
		{
			var ret:Array = [];
			for(var y:int = 0; y < HEIGHT; y++)
			{
				if(IsLineFull(y))
				{
					ret.push(y);
				}
			}

			return ret;
		}

		/**
		 * check if the y-th line is full
		 */
		private function IsLineFull(y:int):Boolean
		{
			for(var x:int = 0; x < WIDTH; x++)
			{
				if(boxes.getAt(x, y) == null)
				{
					return false;
				}
			}

			return true;
		}

		private function setNextTetrimino():void
		{
			tetrimino = Tetrimino.GetRandomTetrimino();
		}
	}
}