package
{
	import flash.filters.*;
	import flash.display.*;
	import flash.text.*;
	import flash.events.*;
	import flash.external.*;
	import flash.geom.*;
	import mx.utils.*;
	
	[SWF(width="300", height="450", backgroundColor="#000000")] 
	public class FieldView extends Sprite
	{
		private var blocks:Array;
		protected var field:Field;
		protected var bmpField:Bitmap;

		protected function get SIZE():int
		{
			return 20;
		}

		public function FieldView()
		{
			stage.scaleMode = "noScale";
			stage.align = "topleft";

			field = new Field(stage);
			field.play();

			// initialize bitmap
			initBlockBitmap();
			initStageBorder();
			addChild(bmpField = new Bitmap());
			bmpField.x = SIZE;
			bmpField.bitmapData = new BitmapData(SIZE * Field.WIDTH, SIZE * Field.HEIGHT, false);

			// event
			stage.addEventListener("enterFrame", function(e:Event):void
			{
				if(field.playing) redraw();
			});
		}

		// initialize block bitmap
		private function initBlockBitmap():void
		{
			if(blocks)
			{
				return;
			}
			blocks = [];

			for each(var k:String in Tetrimino.kinds)
			{
				var bmpdata:BitmapData = new BitmapData(SIZE, SIZE, true, 0);

				drawBlock(bmpdata, k);

				blocks.push(bmpdata);
			}
		}

		// draw block
		protected function drawBlock(bitmap:BitmapData, kind:String):void
		{
			var size:int = SIZE - 1;

			var color:int = Tetrimino.colorList[kind];
			bitmap.fillRect(new Rectangle(0, 0, size, size), color | 0xff000000);
		}

		// draw stage border line
		private function initStageBorder():void
		{
			var block:BitmapData = new BitmapData(SIZE, SIZE, true, 0xff000000);
			drawBlock(block, Tetrimino.WALL);

			var bmp:Bitmap = new Bitmap();
			bmp.bitmapData = new BitmapData(SIZE * (Field.WIDTH + 2), SIZE * (Field.HEIGHT + 1), true, 0);
			addChild(bmp);

			for(var i:int = 0; i < Field.HEIGHT; i++)
			{
				bmp.bitmapData.copyPixels(block, block.rect, new Point(0, i * SIZE));
				bmp.bitmapData.copyPixels(block, block.rect, new Point((Field.WIDTH + 1) * SIZE, i * SIZE));
			}

			for(i = 0; i < Field.WIDTH + 2; i++)
			{
				bmp.bitmapData.copyPixels(block, block.rect, new Point(i * SIZE, Field.HEIGHT * SIZE));
			}
		}

		// update field
		public function redraw():void
		{
			var bmp:BitmapData = bmpField.bitmapData;
			bmp.fillRect(bmp.rect, 0);

			// ̕`
			for(var x:int = 0; x < Field.WIDTH; x++)
			{
				for(var y:int = 0; y < Field.HEIGHT; y++)
				{
					transferBlockToBitmap(field.boxes.getAt(x, y), x, y);
				}
			}

			// tetrimino ̕`
			var t:Tetrimino = field.tetrimino;
			if(t == null) return;
			for each(var box:Object in t.boxes)
			{
				transferBlockToBitmap(t.kind, box.x + t.x, box.y + t.y);
			}
		}

		private function transferBlockToBitmap(data:Object, x:int, y:int):void
		{
			if(data == null || !(data is String))
			{
				return;
			}

			var index:int = Tetrimino.kinds.indexOf(data as String);
			if(index < 0 || index >= blocks.length) return;

			var bmp:BitmapData = bmpField.bitmapData;
			var copiedBmp:BitmapData = blocks[index];
			if(copiedBmp)
			{
				bmp.copyPixels(copiedBmp, copiedBmp.rect, new Point(x * SIZE, y * SIZE));
			}
		}
	}
}