package
{
	import flash.display.*;
	import flash.events.*;
	import flash.geom.*;
	import mx.utils.*;
	import mx.events.*;
	import caurina.transitions.*;
	import org.papervision3d.scenes.*;
	import org.papervision3d.objects.*;
	import org.papervision3d.cameras.*;
	import org.papervision3d.materials.*;
	import org.papervision3d.events.*;
	
	public class FieldView3D extends Sprite
	{
		private var container:Sprite;
		private var mouseIsDown:Boolean;
		private const SIZE:int = 30;

		private var scene:Scene3D;
		private var world:DisplayObject3D;
		private var camera:Camera3D;

		protected var field:Field;
		private var cubeCache:SimpleMatrix;
		private var tetriminoGroup:DisplayObject3D;
		private var prevTetrimino:Object;

		private var baseX:int;
		private var baseY:int;

		public function FieldView3D()
		{
			stage.scaleMode = "noScale";
			stage.align = StageAlign.TOP_LEFT;

			stage.addEventListener(MouseEvent.MOUSE_DOWN, mouseDown);
			stage.addEventListener(MouseEvent.MOUSE_UP, mouseUp);

			field = new Field(stage);
			field.play();

			cubeCache = new SimpleMatrix();

			mouseIsDown = false;
			
			init3D();

			// event
			stage.addEventListener("enterFrame", function(e:Event):void
			{
				if(field.playing) redraw();
			});
		}

		private function setPos(obj:DisplayObject3D, x:int, y:int):Boolean
		{
			var new_x:int = x * SIZE;
			var new_y:int = (-y + Field.HEIGHT / 2) * SIZE;
			var changed:Boolean = false;

			if(new_x != obj.x)
			{
				obj.x = new_x;
				changed = true;
			}

			if(new_y != obj.y)
			{
				obj.y = new_y;
				changed = true;
			}

			return changed;
		}

		private function getColorMaterial(kind:String):ColorMaterial
		{
			var color:int = Tetrimino.colorList[kind];
			var material:ColorMaterial = new ColorMaterial(color, 1);
			material.doubleSided = true;
			material.lineColor = 0;
			material.lineAlpha = 0.5;
			return material;
		}
		
		private function createBlock(kind:String, x:int = SIZE, y:int = SIZE, z:int = SIZE):Cube
		{
			return new Cube(getColorMaterial(kind), x, y, z);
		}

		private function redraw():void
		{
			var changed:Boolean = false;
			
			// ̕`
			var cube:Cube;
			for(var x:int = 0; x < Field.WIDTH; x++)
			{
				for(var y:int = 0; y < Field.HEIGHT; y++)
				{
					var cache:Object = cubeCache.getAt(x, y) || {kind : null};
					var kind:String = field.boxes.getAt(x, y) as String;
					if(kind != cache.kind)
					{
						if(cache.cube)
						{
							world.removeChild(cache.cube);
						}
						if(kind != null)
						{
							cube = createBlock(kind);
							setPos(cube, x, y);
							world.addChild(cube);
						}

						cubeCache.setAt(x, y, {kind : kind, cube : cube});
						changed = true;
					}
				}
			}
			
			// tetrimino ̕`
			var t:Tetrimino = field.tetrimino;
			if(t)
			{
				if(!prevTetrimino
				|| prevTetrimino.kind != t.kind
				|| prevTetrimino.x != t.x
				|| prevTetrimino.y != t.y
				|| prevTetrimino.angle != t.angle)
				{
					if(tetriminoGroup)
					{
						world.removeChild(tetriminoGroup);
						tetriminoGroup = null;
					}

					tetriminoGroup = new DisplayObject3D();
					for(var i:int = 0; i < t.boxes.length; i++)
					{
						cube = createBlock(t.kind);
						setPos(cube, t.boxes[i].x + t.x, t.boxes[i].y + t.y);
						tetriminoGroup.addChild(cube);
					}
					world.addChild(tetriminoGroup);

					prevTetrimino = {};
					prevTetrimino.kind = t.kind
					prevTetrimino.x = t.x;
					prevTetrimino.y = t.y;
					prevTetrimino.angle = t.angle;

					changed = true;
				}
			}
			else
			{
				if(tetriminoGroup)
				{
					world.removeChild(tetriminoGroup);
					prevTetrimino = null;
				}
			}

			if(mouseIsDown)
			{
				moveCamera();
				changed = true;
			}
			if(changed)
			{
				scene.renderCamera(camera);
			}
		}

		private function init3D():void
		{
			container = new Sprite();
			addChild(container);

			container.x = stage.stageWidth/3;
			container.y = stage.stageHeight/2;

			scene = new Scene3D(container);
			world = new DisplayObject3D();
			scene.addChild(world);
			camera = new Camera3D();

			camera.x = 0;
			camera.y = 0;
			camera.focus = 400;
			camera.zoom = 2;

			// ǂ̕`
			for(var i:int = 0; i < Field.HEIGHT; i++)
			{
				setPos(world.addChild(createBlock(Tetrimino.WALL)), -1, i);
				setPos(world.addChild(createBlock(Tetrimino.WALL)), Field.WIDTH, i);
			}
		}

		private function mouseDown(e:Event):void
		{
			mouseIsDown = true;
			baseX = this.container.mouseX;
			baseY = this.container.mouseY;
		}

		private function mouseUp(e:Event):void
		{
			mouseIsDown = false;
		}

		private function moveCamera():void
		{
			if (!mouseIsDown) return;
			
			world.rotationX += (baseY - this.container.mouseY) / 50;
			world.rotationY += (baseX - this.container.mouseX) / 50;
		}
	}
}