package
{
	import flash.display.*;
	import flash.events.*;
	import mx.binding.utils.*;
	import mx.events.*;

	public class Tetrimino
	{
		public const SIZE:int = 30;
		
		/**
		 * kind list
		 */
		public static const kinds:Array = [I, J, L, O, S, T, Z, WALL];

		public static const I:String = "I";
		public static const J:String = "J";
		public static const L:String = "L";
		public static const O:String = "O";
		public static const S:String = "S";
		public static const T:String = "T";
		public static const Z:String = "Z";
		public static const WALL:String = "WALL";

		/**
		 * color list
		 */
		public static const colorList:Object = {
				I : 0x00ffff,
				J : 0x0000ff,
				L : 0xff9900,
				O : 0xffff00,
				S : 0xff0000,
				T : 0xff00ff,
				Z : 0x00ff00,
				WALL : 0x808080
			};

		private var _kind:String;
		private var _color:int;
		
		public var boxes:Array;

		[Bindable]
		public var x:int = 3;

		[Bindable]
		public var y:int = 0;

		[Bindable]
		public var angle:int = 0;

		public function get kind():String
		{
			return _kind;
		}
	
		public function get color():int
		{
			return _color;
		}
		
		/**
		 * constructor
		 */
		public function Tetrimino(_kind:String)
		{
			this._kind = _kind;
			this._color = colorList[_kind];
			angle = 0;

			boxes = new Array();
			initBoxes();

			ChangeWatcher.watch(this, 'angle', function(e:PropertyChangeEvent):void
			{
				var diff:int = (e.newValue as int) - (e.oldValue as int);
				if(diff > 0)
				{
					while(diff > 0)
					{
						internalRotate(true);
						diff--;
					}
				}
				else
				{
					while(diff < 0)
					{
						internalRotate(false);
						diff++;
					}
				}
			});
		}

		/**
		 * Get random tetrimino
		 */
		public static function GetRandomTetrimino():Tetrimino
		{
			var ret:Tetrimino;
			switch(Math.floor(Math.random() * 7))
			{
				case 0:  ret = new Tetrimino(I); break;
				case 1:  ret = new Tetrimino(J); break;
				case 2:  ret = new Tetrimino(L); break;
				case 3:  ret = new Tetrimino(O); break;
				case 4:  ret = new Tetrimino(S); break;
				case 5:  ret = new Tetrimino(T); break;
				case 6: 
				default: ret = new Tetrimino(Z); break;
			}

			ret.angle = Math.floor(Math.random() * 4);
			
			return ret;
		}

		/**
		 * clone
		 */
		public function clone():Tetrimino
		{
			var ret:Tetrimino = new Tetrimino(_kind);
			ret.x = x;
			ret.y = y;
			ret.angle = angle;
			return ret;
		}

		/**
		 * rotate
		 */
		public function rotate(cwr:Boolean):void
		{
			angle = (cwr ? angle + 1 : angle + 3) % 4;
		}

		/**
		 * rotate, but do not change angle.
		 */
		private function internalRotate(cwr:Boolean):void
		{
			var size:int;
			switch(_kind)
			{
				case O:
					return;

				case I:
					size = 3;
					break;

				default:
					size = 2;
					break;
			}

			for each(var box:Object in boxes)
			{
				var c:int = box.x;
				box.x = ( cwr ? size - box.y : box.y);
				box.y = (!cwr ? size -     c :     c);
			}
		}

		/**
		 * init box position
		 */
		private function initBoxes():void
		{
			switch(_kind)
			{
				case I:
					boxes.push({x :  0, y :  1});
					boxes.push({x :  1, y :  1});
					boxes.push({x :  2, y :  1});
					boxes.push({x :  3, y :  1});
					break;

				case J:
					boxes.push({x :  0, y :  0});
					boxes.push({x :  0, y :  1});
					boxes.push({x :  1, y :  1});
					boxes.push({x :  2, y :  1});
					break;

				case L:
					boxes.push({x :  2, y :  0});
					boxes.push({x :  0, y :  1});
					boxes.push({x :  1, y :  1});
					boxes.push({x :  2, y :  1});
					break;

				case O:
					boxes.push({x :  1, y :  1});
					boxes.push({x :  1, y :  2});
					boxes.push({x :  2, y :  1});
					boxes.push({x :  2, y :  2});
					break;

				case S:
					boxes.push({x :  0, y :  1});
					boxes.push({x :  1, y :  1});
					boxes.push({x :  1, y :  0});
					boxes.push({x :  2, y :  0});
					break;

				case T:
					boxes.push({x :  1, y :  0});
					boxes.push({x :  0, y :  1});
					boxes.push({x :  1, y :  1});
					boxes.push({x :  2, y :  1});
					break;

				case Z:
				default:
					boxes.push({x :  0, y :  0});
					boxes.push({x :  1, y :  0});
					boxes.push({x :  1, y :  1});
					boxes.push({x :  2, y :  1});
					break;
			}
		}

		/**
		 * get color from index
		 */
		public static function getColorFromIndex(index:int):int
		{
			if(index >= 0 && index < kinds.length)
			{
				return colorList[kinds[index]];
			}
			return -1;
		}
	}
}